import numpy as np
from typing import List

import supervisely as sly
from supervisely.nn.prediction_dto import PredictionPoint


def numpy_to_dto_point(points: np.ndarray, class_name: str) -> List[PredictionPoint]:
    assert points.shape[-1] == 2
    return [PredictionPoint(class_name=class_name, col=p[1], row=p[0]) for p in points]


def dto_points_to_point_location(points: List[PredictionPoint]) -> List[sly.PointLocation]:
    return [sly.PointLocation(row=p.row, col=p.col) for p in points]


def exteriors_to_sly_polygons(exteriors: List[List[sly.PointLocation]]) -> List[sly.Polygon]:
    return [sly.Polygon(exterior=exterior) for exterior in exteriors]


def dto_points_to_sly_points(points: List[PredictionPoint]) -> List[sly.Point]:
    return [sly.Point(row=p.row, col=p.col) for p in points]



# coding: utf-8

from supervisely.api.entity_annotation.entity_annotation_api import EntityAnnotationAPI
from supervisely.api.module_api import ApiField
from supervisely.pointcloud_annotation.pointcloud_episode_annotation import (
    PointcloudEpisodeAnnotation,
)
from supervisely.video_annotation.key_id_map import KeyIdMap


class PointcloudEpisodeAnnotationAPI(EntityAnnotationAPI):
    """
    :class:`PointcloudEpisodeAnnotation<supervisely.pointcloud_annotation.pointcloud_episode_annotation.PointcloudEpisodeAnnotation>` for a point clouds episodes.
    :class:`PointcloudEpisodeAnnotationAPI<PointcloudEpisodeAnnotationAPI>` object is immutable.

    :param api: API connection to the server.
    :type api: Api
    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
            load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        dataset_id = 62664
        ann_info = api.pointcloud_episode.annotation.download(dataset_id)
        print(ann_info)
    """

    _method_download = "point-clouds.episodes.annotations.info"
    _entity_ids_str = ApiField.POINTCLOUD_IDS

    def download(self, dataset_id: int) -> dict:
        """
        Download information about PointcloudEpisodeAnnotation by dataset ID from API.

        :param dataset_id: :class:`Dataset<supervisely.project.project.Dataset>` ID in Supervisely.
        :type dataset_id: int
        :return: Dictionary woth information about PointcloudEpisodeAnnotation in json format
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from pprint import pprint

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 62664
            ann_info = api.pointcloud_episode.annotation.download(dataset_id)
            pprint(ann_info)

            # Output: {
            #     'datasetId': 62664,
            #     'description': '',
            #     'frames': [
            #         {'figures': [{'classId': None,
            #                     'createdAt': '2023-04-05T08:55:52.445Z',
            #                     'description': '',
            #                     'geometry': {'dimensions': {'x': 2.3652234,
            #                                                 'y': 23.291742,
            #                                                 'z': 3.326648},
            #                                 'position': {'x': 86.29707472161449,
            #                                             'y': -14.472597682830635,
            #                                             'z': 0.8842007608554671},
            #                                 'rotation': {'x': 0,
            #                                             'y': 0,
            #                                             'z': -1.6962800995995606}},
            #                     'geometryType': 'cuboid_3d',
            #                     'id': 87830452,
            #                     'labelerLogin': 'almaz',
            #                     'objectId': 5565741,
            #                     'updatedAt': '2023-04-05T08:55:52.445Z'},
            #                     {'classId': None,
            #                     'createdAt': '2023-04-05T08:55:52.445Z',
            #                     'description': '',
            #                     'geometry': {'indices': [783,
            #                                             784,
            #                                             ...
            #                                             28326,
            #                                             30294]},
            #                     'geometryType': 'point_cloud',
            #                     'id': 87830456,
            #                     'labelerLogin': 'almaz',
            #                     'objectId': 5565740,
            #                     'updatedAt': '2023-04-05T08:55:52.445Z'}],
            #         'index': 0,
            #         'pointCloudId': 19618654},...
            #     ],
            #     'tags': []
            # }
        """

        response = self._api.post(self._method_download, {ApiField.DATASET_ID: dataset_id})
        return response.json()[0]

    def download_bulk(self, dataset_id, entity_ids):
        """Not supported for point cloud episodes."""
        raise RuntimeError("Not supported for episodes")

    def append(
        self,
        dataset_id: int,
        ann: PointcloudEpisodeAnnotation,
        frame_to_pointcloud_ids,
        key_id_map: KeyIdMap = None,
    ) -> None:
        """
        Loads an PointcloudEpisodeAnnotation to a given point cloud ID in the API.

        :param pointcloud_id: Point cloud ID in Supervisely.
        :type pointcloud_id: int
        :param ann: PointcloudEpisodeAnnotation object.
        :type ann: PointcloudEpisodeAnnotation
        :param frame_to_pointcloud_ids:  List of dictionaries with frame_id and name of pointcloud episodes
        :type frame_to_pointcloud_ids: List[dict]
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: None
        :rtype: :class:`NoneType`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            pcd_episodes_id = 198704259
            api.pointcloud_episode.annotation.append(pcd_episodes_id, pointcloud_ann)
        """

        if key_id_map is None:
            # create for internal purposes (to link figures and tags to objects)
            key_id_map = KeyIdMap()

        figures = []
        pointcloud_ids = []
        for i, frame in enumerate(ann.frames):
            for fig in frame.figures:
                if frame_to_pointcloud_ids.get(i) is None:  # skip unmapped frames
                    continue

                figures.append(fig)
                pointcloud_ids.append(frame_to_pointcloud_ids[i])

        self._api.pointcloud_episode.object.append_to_dataset(dataset_id, ann.objects, key_id_map)
        self._api.pointcloud_episode.figure.append_to_dataset(
            dataset_id, figures, pointcloud_ids, key_id_map
        )



import sys
import os
import json
import click
import re
import subprocess
import git
from dotenv import load_dotenv
from rich.console import Console


MIN_SUPPORTED_INSTANCE_VERSION = "6.7.21"

from supervisely.cli.release.release import (
    find_tag_in_repo,
    push_tag,
    get_app_from_instance,
    get_module_root,
    get_module_path,
    get_appKey,
    release,
    slug_is_valid,
    delete_tag,
    get_instance_version,
)


def _check_git(repo: git.Repo):
    console = Console()
    result = True

    if len(repo.untracked_files) > 0:
        console.print(
            "[red][Error][/] You have untracked files. Commit all changes before releasing the app."
        )
        console.print("  Untracked files:")
        console.print("\n".join(
            [f"  {i+1}) " + file for i, file in enumerate(repo.untracked_files)][:20]
        ))
        if len(repo.untracked_files) > 20:
            console.print(f"  ... and {len(repo.untracked_files) - 20} more.")
        print()
        result = False

    if len(repo.index.diff(None)) > 0:
        console.print(
            "[red][Error][/] You have modified files. Commit all changes before releasing the app."
        )
        console.print("  Modified files:")
        console.print(
            "\n\n".join(
                f"  {i+1}) " + str(d).replace("\n", "\n  ")
                for i, d in enumerate(repo.index.diff(None))
            )
        )
        print()
        result = False

    if len(repo.index.diff("HEAD")) > 0:
        console.print(
            "[red][Error][/] You have staged files. Commit all changes before releasing the app."
        )
        console.print("  Staged files:")
        console.print(
            "\n\n".join(
                f"  {i+1}) " + str(d).replace("\n", "\n  ")
                for i, d in enumerate(repo.index.diff("HEAD"))
            )
        )
        print()
        result = False

    local_branch = repo.active_branch
    remote_branch = local_branch.tracking_branch()
    if remote_branch is None:
        console.print(
            "[red][Error][/] Your branch is not tracking any remote branches. Try running 'git push --set-upstream origin'\n"
        )
        result = False
    elif not local_branch.commit.hexsha == remote_branch.commit.hexsha:
        console.print(
            "[red][Error][/] Local branch and remote branch are different. Push your changes to a remote branch before releasing the app\n"
        )
        result = False

    return result


def _ask_release_version(repo: git.Repo):
    console = Console()
    try:
        sly_release_tags = [
            tag.name
            for tag in repo.tags
            if re.match("^sly-release-v\d+\.\d+\.\d+$", tag.name)
        ]
        sly_release_tags.sort(
            key=lambda tag: [
                int(n) for n in tag.split("sly-release-v")[-1].split(".")
            ]
        )
        current_release_version = sly_release_tags[-1][13:]
        suggested_release_version = ".".join(
            [
                *current_release_version.split(".")[:-1],
                str(int(current_release_version.split(".")[-1]) + 1),
            ]
        )
    except Exception:
        current_release_version = None
        suggested_release_version = "0.0.1"
    while True:
        input_msg = f'Enter release version in format vX.X.X ({f"Last release: [blue]{current_release_version}[/]. " if current_release_version else ""}Press "Enter" for [blue]{suggested_release_version}[/]):\n'
        release_version = console.input(input_msg)
        if release_version == "":
            release_version = "v" + suggested_release_version
            break
        if _check_release_version(release_version):
            break
        console.print("Wrong version format. Should be of format vX.X.X")
    
    return release_version


def _ask_release_description():
    console = Console()
    release_description = console.input(
        "Enter release description:\n",
    )
    return release_description


def _ask_confirmation():
    while True:
        confirmed = input("Confirm? [y/n]:\n")
        if confirmed.lower() in ["y", "yes"]:
            return True
        if confirmed.lower() in ["n", "not"]:
            return False


def _check_release_version(release_version):
    return re.fullmatch("v\d+\.\d+\.\d+", release_version)


def _check_instance_version(instance_version):
    last_supported = [int(x) for x in MIN_SUPPORTED_INSTANCE_VERSION.split(".")]
    version_numbers = [int(x) for x in instance_version.split(".")]
    for number, supported in zip(version_numbers, last_supported):
        if number < supported:
            return False
    return True


def run(
    app_directory, sub_app_directory, slug, autoconfirm, release_version, release_description
):
    console = Console()
    console.print("\nSupervisely Release App\n", style="bold")

    # check slug
    if slug and not slug_is_valid(slug):
        console.print("[red][Error][/] Invalid slug")
        return False

    # get module path and check if it is a git repo
    module_root = get_module_root(app_directory)
    module_path = get_module_path(module_root, sub_app_directory)
    try:
        repo = git.Repo(module_root)
    except git.InvalidGitRepositoryError:
        console.print(
            f"[red][Error][/] Module path [green]{module_path}[/] is not a git repository"
        )
        return False

    # get server address
    load_dotenv(os.path.expanduser("~/supervisely.env"))
    server_address = os.getenv("SERVER_ADDRESS", None)
    if server_address is None:
        console.print(
            '[red][Error][/] Cannot find [green]SERVER_ADDRESS[/]. Add it to your "~/supervisely.env" file or to environment variables'
        )
        return False

    # get api token
    api_token = os.getenv("API_TOKEN", None)
    if api_token is None:
        console.print(
            '[red][Error][/] Cannot find [green]API_TOKEN[/]. Add it to your "~/supervisely.env" file or to environment variables'
        )
        return False

    # check instance version
    try:
        instance_version = get_instance_version(api_token, server_address)
        if not _check_instance_version(instance_version["version"]):
            console.print(
                f'[red][Error][/] Instance "{server_address}" does not support releasing apps via sly-release. Please update your instance to version {MIN_SUPPORTED_INSTANCE_VERSION} or higher'
            )
            return False
    except NotImplementedError:
        console.print(
            f'[red][Error][/] Instance "{server_address}" does not support releasing apps via sly-release. Please update your instance to version {MIN_SUPPORTED_INSTANCE_VERSION} or higher'
        )
        return False
    except PermissionError:
        console.print(
            "[red][Error][/] Permission denied. Check that all credentials are set correctly"
        )
        return False

    # get config
    try:
        with open(module_path.joinpath("config.json"), "r") as f:
            config = json.load(f)
    except FileNotFoundError:
        console.print(
            f'[red][Error][/] Cannot find "config.json" file at "{module_path}"'
        )
        return False
    except json.JSONDecodeError as e:
        console.print(
            f'[red][red][Error][/][/] Cannot decode config json file at "{module_path.joinpath("config.json")}": {str(e)}'
        )
        return False
    
    # get readme
    try:
        with open(module_path.joinpath("README.md"), "r", encoding="utf_8") as f:
            readme = f.read()
    except:
        readme = ""

    # get app name
    app_name = config.get("name", None)
    if app_name is None:
        console.print(
            f'[red][Error][/] Missing "name" field in config json file at "{module_path.joinpath("config.json")}"'
        )
        return False

    # get modal template
    modal_template = ""
    if "modal_template" in config:
        modal_template_path = module_root.joinpath(config["modal_template"])
        if not modal_template_path.exists() or not modal_template_path.is_file():
            console.print(
                f'[red][Error][/] Cannot find Modal Template at "{modal_template_path}". Please check your [green]config.json[/] file'
            )
            return False
        with open(modal_template_path, "r") as f:
            modal_template = f.read()

    # print details
    console.print(f"Application directory:\t[green]{module_path}[/]")
    console.print(f"Server address:\t\t[green]{server_address}[/]")
    console.print(f"Api token:\t\t[green]{api_token[:4]}*******{api_token[-4:]}[/]")
    console.print(f"Git branch:\t\t[green]{repo.active_branch}[/]")

    # check that everything is commited and pushed
    success = _check_git(repo)
    if not success:
        return False

    # get appKey
    appKey = get_appKey(repo, sub_app_directory)

    # check if app exist or not
    module_exists_label = "[yellow bold]updated[/]"
    try:
        app_data = get_app_from_instance(appKey, api_token, server_address)
        if app_data is None:
            module_exists_label = "[green bold]created[/]"
    except PermissionError:
        console.print(
            "[red][Error][/] Permission denied. Check that all credentials are set correctly"
        )
        return False

    # get and check release version
    if repo.active_branch.name in ["main", "master"]:
        if release_version is None:
            release_version = _ask_release_version(repo)
        if not _check_release_version(release_version):
            console.print(
                '[red][Error][/] Incorrect release version. Should be of format "vX.X.X"'
            )
            return False
    else:
        console.print(
            f'Release version will be "{repo.active_branch.name}" as the branch name'
        )
        release_version = repo.active_branch.name

    # get release name
    if release_description is None:
        release_description = _ask_release_description()

    # print summary
    console.print(
        f'\nApplication "{app_name}" will be {module_exists_label} at "{server_address}" Supervisely instance with release [blue]{release_version}[/] "{release_description}"'
    )
    if repo.active_branch.name in ["main", "master"]:
        remote_name = repo.active_branch.tracking_branch().name
        console.print(
            f'Git tag "sly-release-{release_version}" will be added and pushed to remote "{remote_name}"'
        )

    # ask for confiramtion if needed
    if autoconfirm:
        confirmed = True
    else:
        confirmed = _ask_confirmation()
    if not confirmed:
        console.print("Release cancelled")
        return False

    # add tag and push
    tag_created = False
    if repo.active_branch.name in ["main", "master"]:
        tag_name = f"sly-release-{release_version}"
        tag = find_tag_in_repo(tag_name, repo)
        if tag is None:
            tag_created = True
            repo.create_tag(tag_name)
        try:
            push_tag(tag_name, repo)
        except subprocess.CalledProcessError:
            if tag_created:
                repo.delete_tag(tag)
            console.print(
                f"[red][Error][/] Git push unsuccessful. You need write permissions in repository to release the application"
            )
            return False

    # release
    console.print("Uploading archive...")
    success = True
    response = release(
        server_address,
        api_token,
        appKey,
        repo,
        config,
        readme,
        release_description,
        release_version,
        modal_template,
        slug,
    )
    if response.status_code != 200:
        error = f"[red][Error][/] Error releasing the application. Please contact Supervisely team. Status Code: {response.status_code}"
        try:
            error += ", " + json.dumps(response.json())
        except:
            pass
        console.print(error)
        success = False
    elif response.json()["success"] != True:
        console.print(
            "[red][Error][/] Error releasing the application. Please contact Supervisely team"
            + json.dumps(response.json())
        )
        success = False

    # delete tag in case of release failure
    try:
        message = response.json()["details"]["message"]
        if message.startswith("version") and message.endswith("already exists"):
            version_exist = True
        else:
            version_exist = False
    except:
        version_exist = False
    if not success and not version_exist:
        if tag_created:
            console.print(f'Deleting tag "sly-release-{release_version}" from remote')
            try:
                delete_tag(f"sly-release-{release_version}", repo)
            except subprocess.CalledProcessError:
                console.print(
                    f'[orange1][Warning][/] Could not delete tag "sly-release-{release_version}" from remote. Please do it manually to avoid errors'
                )
        return False

    return True


@click.command(
    help="This app allows you to release your aplication to Supervisely platform"
)
@click.option(
    "-p",
    "--path",
    required=False,
    help="[Optional] Path to the directory with application",
)
@click.option(
    "-a",
    "--sub-app",
    required=False,
    help="[Optional] Path to sub-app relative to application directory",
)
@click.option(
    "--release-version",
    required=False,
    help='[Optional] Release version in format "vX.X.X"',
)
@click.option("--release-description", required=False, help="[Optional] Release description")
@click.option("-y", is_flag=True, help="[Optional] Add this flag for autoconfirm")
@click.option("-s", "--slug", required=False, help="[Optional] For internal use")
def cli_run(path, sub_app, slug, y, release_version, release_description):
    try:
        success = run(
            app_directory=path,
            sub_app_directory=sub_app,
            slug=slug,
            autoconfirm=y,
            release_version=release_version,
            release_description=release_description,
        )
        if success:
            print("App released sucessfully!")
            sys.exit(0)
        else:
            print("App not released")
            sys.exit(1)
    except KeyboardInterrupt:
        print("Aborting...")
        print("App not released")
        sys.exit(1)



from __future__ import annotations
from collections import namedtuple
from supervisely.app.widgets import Widget
from supervisely.app.content import DataJson, StateJson

from typing import List, Dict, Union, Optional, Callable

class Transfer(Widget):
    """Widget for transfering items between source and target lists.
    Important info: the widget operates with a single list of items (not with two lists of items).
    So, when initializing widget, you should pass a single list of items, which will be displayed in both lists.
    If you want to initialize widget with some items in the right (target) list, you can pass them in the
    `transferred_items` argument, which should containt keys of the items in the 'items' list, NOT NEW items.
    Otherwise, they won't be displayed. Likewise, if you want to change the current list of transferred items
    (which shown in the right list), you can use the 'set_transferred_items' method, which works exactly
    the same as the argument 'transferred_items' in the constructor:
    you should pass keys of the items in the 'items' list.

    :param items: A list of Transfer.Item objects or a list of strings (keys), which will be automatically
                transformed into Transfer.Item objects. Defaults to None.
    :type items: Union[List[Transfer.Item], List[str]], optional
    :param transferred_items: A list of keys of the items to be displayed in the right (target) list. This list
                            contains keys of the items in the 'items' list, not new items. Defaults to None.
    :type transferred_items: List[str], optional
    :param widget_id: The id of the widget. Defaults to None.
    :type widget_id: str, optional
    :param filterable: If True, the widget will have a filter input. Defaults to False.
    :type filterable: bool, optional
    :param filter_placeholder: The placeholder for the filter input. This placeholder will be displayed only if
                            the filterable argument is True. Defaults to None.
    :type filter_placeholder: str, optional
    :param titles: A list of titles for the source and target lists. If not specified, the default titles will be
                used ('Source' and 'Target'). Defaults to None.
    :type titles: List[str], optional
    :param button_texts: A list of texts for the buttons, which transfer items between the lists. If not
                        specified, the buttons will have only icons (> and <). Defaults to None.
    :type button_texts: List[str], optional
    :param left_checked: A list of keys of the items in the left (source) list, which should be checked at widget
                        initialization. Defaults to None.
    :type left_checked: List[str], optional
    :param right_checked: A list of keys of the items in the right (target) list, which should be checked at widget
                        initialization. Defaults to None.
    :type right_checked: List[str], optional
            
    :Methods:
    get_transferred_items(): returns the list of keys of the items, which are currently displayed in the
        right (target) list.
    get_untransferred_items(): returns the list of keys of the items, which are currently displayed in the
        left (source) list.
    @value_changed(func): decorator function, which can be used to handle the events, when the list of
        transferred items is changed (transferred items are moved from the left to the right list or vice versa).
        Function under this decorator should recieve one argument - the list of transferred items (keys of the
        items which are currently displayed in the right list).
    set_items(): sets the list of items to be displayed in the widget. Likewise in the class constructor,
        the list of items can either be a list of Transfer.Item objects or a list of strings.
    set_transferred_items(): sets the list of transferred items (keys of the items which should be displayed
        in the right list).
    add(): adds new items to the widget. Likewise in the class constructor, the list of items can either be
        a list of Transfer.Item objects or a list of strings. Must not contain items with the same keys as
        in the current list of items.
    remove(): removes items from the widget by their keys.
    get_items_keys(): returns the list of keys of the items, which are currently displayed in the widget.

    :Usage example:

     .. code-block:: python
        from supervisely.app.widgets import Transfer

        # Creating widget items with Transfer.Item objects.
        item1 = Transfer.Item(key="cat", label="Cat", disabled=True)
        item2 = Transfer.Item(key="dog", label="Dog")

        # Creating Transfer widget with the list of Transfer.Item objects. The item "dog" will be displayed
        # in the right (target) list at widget initialization.
        transfer = Transfer(items=[item1, item2], transferred_items=["dog"])

        # Setting new transferred items. The item "cat" will be displayed in the right (target) list.
        # Note: items that was in transferred_items list before, but not in the new list, will be moved
        # to untransferred items list (left).
        transfer.set_transferred_items(["cat"])
        print(transfer.get_transferred_items()) # ["cat"]
        print(transfer.get_untransferred_items()) # ["dog"]

        # Creating empty Transfer widget.
        transfer = Transfer()
        # Adding items (as strings) to the widget. The item "dog" will be displayed in the right (target) list.
        transfer.set(["cat", "dog", "mouse"], transferred_items=["dog"])
    """
    class Routes:
        VALUE_CHANGED = "value_changed"
        
    class Item:
        """
        Class for representing items in the Transfer widget.

        Each item has a key, label and disabled flag. Key is required and should be unique. Label is optional and will be
        displayed in the widget. If not specified, label will be equal to the key. Disabled flag is optional and if True,
        the item will be disabled and won't be transferable. Disabled flag is False by default.

        :param key: The key of the item which is mostly used for identifying the item (like value).
            It's important to note, that the keys of the items should be unique.
        :type key: str
        :param label: The label of the item, which will be displayed in the widget. If not specified, the key will be used
            as label. Defaults to None.
        :type label: Optional[str]
        :param disabled: If True, the item will be disabled and won't be transferable. Defaults to False.
        :type disabled: Optional[bool]

        :Usage example:

         .. code-block:: python

            disabled_item = Transfer.Item(key="dog", label="Dog", disabled=True)
            item_with_label = Transfer.Item(key="cat", label="Cat")
            simple_item = Transfer.Item(key="mouse")
        """

        def __init__(self, key: str, label: Optional[str] = None, disabled: Optional[bool] = False):
            self.key = key
            if not label:
                # If label is not specified, the key will be used as label.
                self.label = key
            else:
                self.label = label
            self.disabled = disabled
    
        def to_json(self):
            return {"key": self.key, "label": self.label, "disabled": self.disabled}
    
    def __init__(self, 
                 items: Optional[Union[List[Item], List[str]]] = None, 
                 transferred_items: Optional[List[str]] = None, 
                 widget_id: Optional[str] = None,
                 filterable: Optional[bool] = False, 
                 filter_placeholder: Optional[str] = None, 
                 titles: Optional[List[str]] = None,
                 button_texts: Optional[List[str]] = None, 
                 left_checked: Optional[List[str]] = None, 
                 right_checked: Optional[List[str]] = None):    
        
        self._changes_handled = False
        self._items = None
        self._transferred_items = []
        
        if items:
            self._items = self.__checked_items(items)
        
        if transferred_items:
            self._transferred_items = self.__checked_transferred_items(transferred_items)
        
        # If wrong items are specified, items won't be checked.
        self._left_checked = left_checked
        self._right_checked = right_checked
        
        self._filterable = filterable
        self._filter_placeholder = filter_placeholder
        
        self._titles = titles if titles is not None else ["Source", "Target"]
        
        self._button_texts = button_texts
        
        super().__init__(widget_id=widget_id, file_path=__file__)
    
    
    def __checked_items(self, items: Optional[Union[List[Item], List[str]]]) -> List[Transfer.Item]:
        """
        If the list of items is specified as a list of strings, they will be converted to Transfer.Item objects. List of
        Transfer items will be checked for uniqueness of the keys. If the keys of the items are not unique, an error will be
        raised.

        :param items: The list of items can either be a list of Transfer.Item objects or a list of strings, containing the
            keys for Transfer.Item objects to be created.
        :type items: Optional[Union[List[Item], List[str]]]

        :raises ValueError: If the keys of the items are not unique.

        :return: The list of Transfer.Item objects with unique keys.
        :rtype: List[Transfer.Item]
        """

        if isinstance(items[0], str):
            # If items are specified as strings, they will be converted to Transfer.Item objects.
            if len(set(items)) != len(items):
                # If the keys of the items are not unique, an error will be raised.
                raise ValueError("The keys of the items should be unique.")
            
            checked_items = [Transfer.Item(key=item) for item in items]
        else:
            # If items are specified as Transfer.Item objects, they will be checked for uniqueness.
            if len({item.key for item in items}) != len(items):
                # If the keys of the items are not unique, an error will be raised.
                raise ValueError("The keys of the items should be unique.")
            
            checked_items = items
        return checked_items
    
    def __checked_transferred_items(self, transferred_items: List[str]) -> List[str]:
        """
        If the self._items is specified, the list of transferred items will be checked for the keys of the items. Since
        transferred items are specified by keys of the items, each key of the transferred items should exist in the list of
        items. Otherwise, an error will be raised.

        :param transferred_items: List of keys of the items to be shown in the right (target) list.
        :type transferred_items: List[str]

        :raises ValueError: If transferred items are specified, but the list of items is not specified.
        :raises ValueError: If any of transferred items keys is not in the list of items.

        :return: List of transferred items (keys of the items which should be displayed in the right list).
        :rtype: List[str]
        """

        if not self._items:
            # If the list of items is not specified, the list of transferred items can't be specified since
            # the keys of the items should be specified in the list of transferred items.
            raise ValueError("The 'items' argument should be specified if "\
                "the 'transferred_items' argument is specified.")
        else:
            if not set(transferred_items).issubset([item.key for item in self._items]):
                # If any of the keys of the items specified in the list of transferred items is not
                # in the list of items, an error will be raised.
                raise ValueError("The 'transferred_items' argument should contain only "\
                    "the keys of the items specified in the 'items' argument.")
            else:
                return transferred_items
    
    def get_json_data(self) -> Dict[str, Union[List[Dict[str, Union[str, bool]]], None]]:
        """
        Returns the data of the widget in JSON format.

        Data will contain the list of items and the list of transferred items.

        :return: The data of the widget in JSON format: {"items": List[Dict[str, Any]], "transferred_items": List[str]}.
            "items" - the list of items in the widget in JSON format. Each item is represented as Transfer.Item object.
            "transferred_items" - the list of transferred items (keys of the items which should be displayed in the right list).
        :rtype: Dict[str, List[Dict[str, Union[str, bool]]]]
        """

        res = {
            "items": None,
        }
        if self._items is not None:
            res["items"] = [item.to_json() for item in self._items]

        return res
    
    def get_json_state(self) -> Dict[str, List[str]]:
        """
        Returns the state of the widget in JSON format.

        State will contain the list of transferred items.

        :return: The state of the widget in JSON format: {"transferred_items": List[str]}. "transferred_items" - the list of
            transferred items (keys of the items which should be displayed in the right list).
        :rtype: Dict[str, List[str]]
        """

        transferred_items = self._transferred_items
        
        return {"transferred_items": transferred_items}
    
    def get_transferred_items(self) -> List[str]:        
        """
        Returns the list of transferred items.

        :return: List of transferred items (keys of the items which should be displayed in the right list).
        :rtype: List[str]
        """

        return StateJson()[self.widget_id]["transferred_items"]
    
    
    def get_untransferred_items(self) -> List[str]:
        """
        Returns the list of untransferred items.

        :return: List of untransferred items (keys of the items which should be displayed in the left list).
        :rtype: List[str]
        """

        return [item.key for item in self._items if item.key not in self.get_transferred_items()]
    
    
    def value_changed(self, func: Callable) -> Callable:
        """
        Decorates a function which will be called when the the items in right list are changed (moved in or out of the list).

        :param func: Function to be wrapped with the decorator. The function should have one argument which will contain
            namedtuple with the following fields: transferred_items, untransferred_items.
        :type func: Callable

        :return: Wrapped function.
        :rtype: Callable
        
        :Usage example:

         .. code-block:: python

            tr = Transfer(items=["item1", "item2", "item3"], transferred_items=["item1"])

            # Move "item2" from the left list to the right list. 
            @tr.value_changed
            def func(items):        
                print(items.transferred_items) # ["item1", "item2"]
                print(items.untransferred_items) # ["item3"]
        """

        
        route_path = self.get_route_path(Transfer.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            
            Items = namedtuple("Items", ["transferred_items", "untransferred_items"])
            
            res = Items(transferred_items=self.get_transferred_items(), untransferred_items=self.get_untransferred_items())
            
            func(res)

        return _click
    
    def set_items(self, items: Union[List[Transfer.Item], List[str]]):
        """
        Sets the list of items for the widget.

        If the list of items is specified as strings, they will be converted to Transfer.Item objects.
        Note: this method will REPLACE the current list of items with the new one. If you want to add new items to the current
        list, use .add() method.

        :param items: List of items to be set as the new list of items for the widget. Can be specified as Transfer.Item
            objects or as strings of the item keys.
        :type items: Union[List[Transfer.Item], List[str]]

        :Usage example:

         .. code-block:: python

            tr = Transfer(items=["cat", "dog"])
            print(tr.get_untransferred_items()) # ["cat", "dog"]
                
            tr.set(items=["bird", "mouse"])
            print(tr.get_untransferred_items()) # ["bird", "mouse"]

            # As you can see, the list of items was replaced with the new one.
        """

        self._items = self.__checked_items(items)

        self.update_data()
        DataJson().send_changes()

    def set_transferred_items(self, transferred_items: List[str]):
        """
        Sets the list of transferred items.
        The list should contain only the keys of the items specified in the list of items. Otherwise, an error will be raised.
        :param transferred_items: List of keys of the items which should be displayed in the right list.
        :type transferred_items: List[str]
        """

        self._transferred_items = self.__checked_transferred_items(transferred_items)
        self.update_state()
        StateJson().send_changes()
        
    def add(self, items: Union[List[Item], List[str]]):
        """
        Adds new items to the current list of items.

        If the list of items is specified as strings, Transfer.Item objects will be created from them. If the list of adding 
        items contains any items with the same key as the items in the current list, an error will be raised.

        :param items: List of items to be added to the current list of items. Can be specified as Transfer.Item objects or
            as strings of the item keys.
        :type items: Union[List[Item], List[str]]

        :raises ValueError: If the list of adding items contains any items with the same key as the items in the current list.

        :Usage example:

         .. code-block:: python

            tr = Transfer(items=["cat", "dog"])
            print(tr.get_untransferred_items()) # ["cat", "dog"]
                
            tr.add(items=["bird", "mouse"])
            print(tr.get_untransferred_items()) # ["cat", "dog", "bird", "mouse"]
        """

        items = self.__checked_items(items)
        
        if any([item.key in [item.key for item in self._items] for item in items]):
            raise ValueError("The 'items' argument should not contain any items with the same key as the items in the current list.")
        else:
            self._items.extend(items)
            self.update_data()
            DataJson().send_changes()
            
    def remove(self, items_keys: List[str]):
        """
        Removes items from the current list of items.

        The list of items to be removed should contain keys of the items which should be removed. If there are no items with 
        the specified keys in the current list, nothing will be removed and no error will be raised.

        :param items_keys: List of keys of the items to be removed.
        :type items_keys: List[str]
        """

        self._items = [item for item in self._items if item.key not in items_keys]
        self._transferred_items = [item for item in self._transferred_items if item not in items_keys]
        self.update_data()
        self.update_state()
        DataJson().send_changes()
        StateJson().send_changes()
    
    def get_items_keys(self) -> List[str]:
        """
        Returns the list of keys of the items.

        :return: List of keys of the items.
        :rtype: List[str]
        """

        return [item.key for item in self._items]



# coding: utf-8
from __future__ import annotations
from typing import List, NamedTuple, Dict, Optional

from supervisely.api.module_api import ApiField, ModuleApi, RemoveableBulkModuleApi
from supervisely.video_annotation.key_id_map import KeyIdMap


class ObjectApi(RemoveableBulkModuleApi):
    """
    Object for :class:`VideoAnnotation<supervisely.video_annotation.video_annotation.VideoAnnotation>`.
    """
    @staticmethod
    def info_sequence():
        """
        NamedTuple ObjectInfo information about Object.

        :Example:

         .. code-block:: python

            ObjectInfo(id=152118,
                       description='',
                       created_at='2021-03-23T13:25:34.705Z',
                       updated_at='2021-03-23T13:25:34.705Z',
                       dataset_id=466642, class_id=2856942,
                       entity_id=198703211,
                       tags=[{'objectId': 152118, 'tagId': 29098694, 'entityId': None, 'id': 40632, 'value': 'grey'}],
                       meta={},
                       created_by_id=16154)
        """
        return [ApiField.ID,
                ApiField.DESCRIPTION,
                ApiField.CREATED_AT,
                ApiField.UPDATED_AT,
                ApiField.DATASET_ID,
                ApiField.CLASS_ID,
                ApiField.ENTITY_ID,
                ApiField.TAGS,
                ApiField.META,
                ApiField.CREATED_BY_ID,
                ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.video.object.info_tuple_name()
            print(tuple_name) # ObjectInfo
        """

        return 'ObjectInfo'

    def get_info_by_id(self, id: int) -> NamedTuple:
        """
        Get Object information by ID.

        :param id: Object ID in Supervisely.
        :type id: int
        :return: Information about Object. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`NamedTuple`
        """
        return self._get_info_by_id(id, 'objects.info')

    def get_list(self, dataset_id: int, filters: Optional[List[Dict[str, str]]]=None) -> List[NamedTuple]:
        """
        Get list of information about all video Objects for a given dataset ID.

        :param dataset_id: Dataset ID in Supervisely.
        :type dataset_id: int
        :param filters: List of parameters to sort output Objects.
        :type filters: List[dict], optional
        :return: Information about Objects. See :class:`info_sequence<info_sequence>`
        :rtype: :class:`List[NamedTuple]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervise.ly'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            dataset_id = 466642
            object_infos = api.video.object.get_list(dataset_id)
            print(object_infos)
            # Output: [
            #     [
            #         152118,
            #         "",
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         466642,
            #         2856942,
            #         198703211,
            #         [
            #             {
            #                 "objectId": 152118,
            #                 "tagId": 29098694,
            #                 "entityId": null,
            #                 "id": 40632,
            #                 "value": "grey"
            #             }
            #         ],
            #         {},
            #         16154
            #     ],
            #     [
            #         152119,
            #         "",
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         466642,
            #         2856942,
            #         198703211,
            #         [
            #             {
            #                 "objectId": 152119,
            #                 "tagId": 29098694,
            #                 "entityId": null,
            #                 "id": 40633,
            #                 "value": "wine"
            #             }
            #         ],
            #         {},
            #         16154
            #     ],
            #     [
            #         152120,
            #         "",
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         466642,
            #         2856942,
            #         198703211,
            #         [
            #             {
            #                 "objectId": 152120,
            #                 "tagId": 29098694,
            #                 "entityId": null,
            #                 "id": 40634,
            #                 "value": "beige"
            #             }
            #         ],
            #         {},
            #         16154
            #     ],
            #     [
            #         152121,
            #         "",
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         466642,
            #         2856941,
            #         198703212,
            #         [
            #             {
            #                 "objectId": 152121,
            #                 "tagId": 29098696,
            #                 "entityId": null,
            #                 "id": 40635,
            #                 "value": "juvenile"
            #             }
            #         ],
            #         {},
            #         16154
            #     ],
            #     [
            #         152122,
            #         "",
            #         "2021-03-23T13:25:34.705Z",
            #         "2021-03-23T13:25:34.705Z",
            #         466642,
            #         2856943,
            #         198703211,
            #         [],
            #         {},
            #         16154
            #     ]
            # ]
        """
        return self.get_list_all_pages('annotation-objects.list',  {ApiField.DATASET_ID: dataset_id,
                                                                    ApiField.FILTER: filters or []})

    def _append_bulk(self, tag_api, entity_id, project_id, dataset_id, objects, key_id_map: KeyIdMap = None, is_pointcloud=False):
        """"""
        if len(objects) == 0:
            return []

        objcls_name_id_map = self._api.object_class.get_name_to_id_map(project_id)

        items = []
        for obj in objects:
            new_obj = {ApiField.CLASS_ID: objcls_name_id_map[obj.obj_class.name]}

            if not is_pointcloud:
            #if entity_id is not None:
                new_obj[ApiField.ENTITY_ID] = entity_id
            items.append(new_obj)

        response = self._api.post('annotation-objects.bulk.add', {ApiField.DATASET_ID: dataset_id,
                                                                  ApiField.ANNOTATION_OBJECTS: items})
        ids = [obj[ApiField.ID] for obj in response.json()]
        KeyIdMap.add_objects_to(key_id_map, [obj.key() for obj in objects], ids)

        # add tags to objects
        tag_api.append_to_objects(entity_id, project_id, objects, key_id_map)

        return ids



import numpy as np
from typing import Optional, List

import supervisely as sly
from supervisely.geometry.geometry import Geometry


class TrackerInterface:
    def __init__(self, context, api):
        self.api: sly.Api = api
        self.logger = api.logger
        self.frame_index = context["frameIndex"]
        self.frames_count = context["frames"]

        self.track_id = context["trackId"]
        self.video_id = context["videoId"]
        self.object_ids = list(context["objectIds"])
        self.figure_ids = list(context["figureIds"])
        self.direction = context["direction"]
        self.stop = len(self.object_ids) * self.frames_count
        self.global_pos = 0
        self.global_stop_indicatior = False

        self.geometries: List[Geometry] = []
        self.frames_indexes: List[int] = []
        self._frames: Optional[np.ndarray] = None

        self._add_geometries()
        self._add_frames_indexes()
        self._load_frames()

    def add_object_geometries(self, geometries: List[Geometry], object_id: int):
        for frame, geometry in zip(self.frames_indexes[1:], geometries):
            self.api.video.figure.create(
                self.video_id,
                object_id,
                frame,
                geometry.to_json(),
                geometry.geometry_name(),
                self.track_id,
            )
            self._notify()

            if self.global_stop_indicatior:
                self.logger.info("Task stoped by user.")
                self._notify(True)
                break

    def _add_geometries(self):
        for figure_id in self.figure_ids:
            figure = self.api.video.figure.get_info_by_id(figure_id)
            geometry = sly.deserialize_geometry(figure.geometry_type, figure.geometry)
            self.geometries.append(geometry)

            if isinstance(geometry, sly.Point):
                self.stop += 1
            elif isinstance(geometry, sly.Polygon):
                self.stop += len(geometry.exterior) + len(geometry.interior)

            # TODO: other geometries

    def _add_frames_indexes(self):
        total_frames = self.api.video.get_info_by_id(self.video_id).frames_count
        cur_index = self.frame_index

        while (
            0 <= cur_index < total_frames
            and len(self.frames_indexes) < self.frames_count + 1
        ):
            self.frames_indexes.append(cur_index)
            cur_index += 1 if self.direction == "forward" else -1

    def _load_frames(self):
        rgbs = []
        for frame_index in self.frames_indexes:
            img_rgb = self.api.video.frame.download_np(self.video_id, frame_index)
            rgbs.append(img_rgb)
        self._frames = rgbs

    def _notify(self, stop: bool = False):
        self.global_pos += 1
        if stop:
            pos = self.stop
        else:
            pos = self.global_pos

        self.logger.debug(f"Notification status: {pos}/{self.stop}")

        self.global_stop_indicatior = self.api.video.notify_progress(
            self.track_id,
            self.video_id,
            min(self.frames_indexes),
            max(self.frames_indexes),
            pos,
            self.stop,
        )

        self.logger.debug(f"Notification status: stop={self.global_stop_indicatior}")

    @property
    def frames(self) -> np.ndarray:
        return self._frames

    @property
    def frames_with_notification(self) -> np.ndarray:
        """Use this in prediction."""
        self._notify()
        return self._frames



import numpy as np
import functools
from fastapi import Request, BackgroundTasks
from typing import Any, Dict, List, Literal, Optional, Union
from typing_extensions import Literal

import supervisely as sly
from supervisely.geometry.geometry import Geometry
from supervisely.nn.prediction_dto import Prediction, PredictionPoint
from supervisely.nn.inference.tracking.tracker_interface import TrackerInterface
from supervisely.nn.inference import Inference
import supervisely.nn.inference.tracking.functional as F


class PointTracking(Inference):
    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        super().__init__(
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def serve(self):
        super().serve()
        server = self._app.get_server()

        @server.post("/track")
        def start_track(request: Request, task: BackgroundTasks):
            task.add_task(track, request)
            return {"message": "Track task started."}

        def send_error_data(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                value = None
                try:
                    value = func(*args, **kwargs)
                except Exception as e:
                    request: Request = args[0]
                    context = request.state.context
                    api: sly.Api = request.state.api
                    track_id = context["trackId"]

                    api.post(
                        "videos.notify-annotation-tool",
                        data={
                            "type": "videos:tracking-error",
                            "data": {
                                "trackId": track_id,
                                "error": {"message": repr(e)},
                            },
                        },
                    )
                return value

            return wrapper

        @send_error_data
        def track(request: Request):
            context = request.state.context
            api: sly.Api = request.state.api

            video_interface = TrackerInterface(
                context=context,
                api=api,
            )
            api.logger.info("Start tracking.")

            for geom, obj_id in zip(video_interface.geometries, video_interface.object_ids):
                if isinstance(geom, sly.Point):
                    geometries = self._predict_point_geometries(geom, video_interface)
                elif isinstance(geom, sly.Polygon):
                    if len(geom.interior) > 0:
                        raise ValueError(f" Can't track polygons with iterior.")
                    geometries = self._predict_polygon_geometries(geom, video_interface)
                else:
                    raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                video_interface.add_object_geometries(geometries, obj_id)
                api.logger.info(f"Object #{obj_id} tracked.")

    def predict(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_object: PredictionPoint,
    ) -> List[PredictionPoint]:
        """
        Track point on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_objects: point to track on the initial frame
        :type start_objects: PredictionPoint
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[PredictionPoint]
        """
        raise NotImplementedError

    def _get_obj_class_shape(self):
        return sly.Point

    def _predict_point_geometries(
        self,
        geom: sly.Point,
        interface: TrackerInterface,
    ) -> List[Geometry]:
        pp_geom = PredictionPoint("point", col=geom.col, row=geom.row)
        predicted: List[Prediction] = self.predict(
            interface.frames_with_notification,
            self.custom_inference_settings_dict,
            pp_geom,
        )
        return F.dto_points_to_sly_points(predicted)

    def _predict_polygon_geometries(
        self,
        geom: sly.Polygon,
        interface: TrackerInterface,
    ) -> List[Geometry]:
        polygon_points = F.numpy_to_dto_point(geom.exterior_np, "polygon")
        exterior_per_time = [[] for _ in range(interface.frames_count)]

        for pp_geom in polygon_points:
            points: List[Prediction] = self.predict(
                interface.frames_with_notification,
                self.custom_inference_settings_dict,
                pp_geom,
            )
            points_loc = F.dto_points_to_point_location(points)
            for fi, point_loc in enumerate(points_loc):
                exterior_per_time[fi].append(point_loc)

        return F.exteriors_to_sly_polygons(exterior_per_time)



# coding: utf-8

import uuid
from supervisely.annotation.tag import Tag, TagJsonFields
from supervisely._utils import take_with_default
from supervisely.volume_annotation.constants import KEY, ID
from supervisely.video_annotation.key_id_map import KeyIdMap


class VolumeTag(Tag):
    """
    VolumeTag object for :class:`VolumeAnnotation<supervisely.volume_annotation.volume_annotation.VolumeAnnotation>`. :class:`VolumeTag<VolumeTag>` object is immutable.

    :param meta: General information about Volume Tag.
    :type meta: TagMeta
    :param value: Volume Tag value. Depends on :class:`TagValueType<TagValueType>` of :class:`TagMeta<TagMeta>`.
    :type value: Optional[Union[str, int, float]]
    :param key: uuid.UUID object.
    :type key: uuid.UUID, optional
    :param sly_id: Volume Tag ID in Supervisely.
    :type sly_id: int, optional
    :param labeler_login: Login of user who created VolumeTag.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when VolumeTag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when VolumeTag was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.volume_annotation.volume_tag import VolumeTag

        meta_brain = sly.TagMeta('brain', sly.TagValueType.NONE)
        # Now we can create a VolumeTag using our TagMeta
        tag_brain = VolumeTag(meta_brain)
        # When you are creating a new Tag
        # Tag.value is automatically cross-checked against your TagMeta value type to make sure the value is valid.
        # If we now try to add a value to our newly created Tag, we receive "ValueError", because our TagMeta value type is "NONE"
        tag_dog = VolumeTag(meta_brain, value="Brain")
        # Output: ValueError: Tag brain can not have value Husky

        # Let's create another Tag with a string value type
        meta_heart = sly.TagMeta('heart', sly.TagValueType.ANY_STRING)
        tag_heart = VolumeTag(meta_heart, value="Heart")

        # Now let's create a Tag using TagMeta with "ONEOF_STRING" value type
        # In order to use "oneof_string value type", you must initialize a variable with possible values(see class TagMeta for more information)
        colors = ["brown", "white", "black", "red", "chocolate", "gold", "grey"]
        meta_lang_color = sly.TagMeta('lang color', sly.TagValueType.ONEOF_STRING, possible_values=colors)
        tag_lang_color = VolumeTag(meta_lang_color, value="white")

        # If given value is not in a list of possible Tags, ValueError will be raised
        tag_lang_color = VolumeTag(meta_lang_color, value="yellow")
        # Output: ValueError: Tag lang color can not have value yellow
    """

    def __init__(
        self,
        meta,
        value=None,
        key=None,
        sly_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        super(VolumeTag, self).__init__(
            meta,
            value=value,
            sly_id=sly_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        self._key = take_with_default(key, uuid.uuid4())

    def key(self) -> str:
        """
        Get key value.

        :return: Get key value.
        :rtype: str
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.volume_annotation.volume_tag import VolumeTag

            meta_heart = sly.TagMeta('heart', sly.TagValueType.NONE)
            tag_heart = VolumeTag(meta_heart)

            print(tag_heart.key()) 
            # Output: 2d9cd2cd-f89c-40a9-8675-1b0773fa250d
        """
        return self._key

    def to_json(self, key_id_map: KeyIdMap = None):
        """
        Convert the VolumeTag to a json dict. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param key_id_map: Key ID Map object.
        :type key_id_map: KeyIdMap, optional
        :return: Json format as a dict
        :rtype: :class:`dict`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.volume_annotation.volume_tag import VolumeTag
            meta_heart = sly.TagMeta('heart', sly.TagValueType.NONE)
            tag_heart = VolumeTag(meta_heart)

            tag_heart_json = tag_heart.to_json()

            print(tag_heart_json)
            # Output: {
            #     "name": "heart",
            #     "key": "058ad7993a534082b4d94cc52542a97d"
            # }
        """

        data_json = super(VolumeTag, self).to_json()
        if type(data_json) is str:
            # @TODO: case when tag has no value, super.to_json() returns tag name
            data_json = {TagJsonFields.TAG_NAME: data_json}
        data_json[KEY] = self.key().hex

        if key_id_map is not None:
            item_id = key_id_map.get_tag_id(self.key())
            if item_id is not None:
                data_json[ID] = item_id

        return data_json

    @classmethod
    def from_json(cls, data: dict, tag_meta_collection, key_id_map: KeyIdMap = None):
        """
        Convert a json dict to VolumeTag. Read more about `Supervisely format <https://docs.supervise.ly/data-organization/00_ann_format_navi>`_.

        :param data: VolumeTag in json format as a dict.
        :type data: dict
        :param tag_meta_collection: :class:`TagMetaCollection<supervisely.annotation.tag_meta_collection.TagMetaCollection>` object.
        :type tag_meta_collection: TagMetaCollection
        :param key_id_map: Key ID Map object.
        :type key_id_map: KeyIdMap, optional
        :return: VolumeTag object
        :rtype: :class:`VolumeTag<VolumeTag>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            tag_heart_json = {
                "name": "heart",
                "value": "777"
            }

            from supervisely.volumme_annotation.volume_tag import VolumeTag
            meta_heart = sly.TagMeta('heart', sly.TagValueType.ANY_STRING)
            meta_collection = sly.TagMetaCollection([meta_heart])

            tag_heart = VolumeTag.from_json(tag_heart_json, meta_collection)
        """

        temp = super(VolumeTag, cls).from_json(data, tag_meta_collection)
        key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_tag(key, data.get(ID, None))

        return cls(
            meta=temp.meta,
            value=temp.value,
            key=key,
            sly_id=temp.sly_id,
            labeler_login=temp.labeler_login,
            updated_at=temp.updated_at,
            created_at=temp.created_at,
        )

    def clone(
        self,
        meta=None,
        value=None,
        key=None,
        sly_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        """
        Makes a copy of VolumeTag with new fields, if fields are given, otherwise it will use fields of the original VolumeTag.

        :param meta: General information about VolumeTag.
        :type meta: VolumeTag, optional
        :param value: VolumeTag value. Depends on :class:`TagValueType<TagValueType>` of :class:`TagMeta<TagMeta>`.
        :type value: str or int or float or None, optional
        :param key: uuid.UUID object.
        :type key: uuid.UUID, optional
        :param sly_id: VolumeTag ID in Supervisely.
        :type sly_id: int, optional
        :param labeler_login: Login of user who created VolumeTag.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when VolumeTag was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when VolumeTag was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        :Usage example:

         .. code-block:: python

            import supervisely as sly
            from supervisely.volume_annotation.volume_tag import VolumeTag

            meta_heart = sly.TagMeta('heart_tag', sly.TagValueType.ANY_STRING)
            heart_tag = VolumeTag(meta_heart, value='Heart')

            meta_heart_2 = sly.TagMeta('heart tag 2', sly.TagValueType.ANY_STRING)

            new_tag = car_tag.clone(meta=meta_heart_2, key=car_tag.key())
            new_tag_json = new_tag.to_json()

            print(new_tag_json)
            # Output: {
            #     "name": "heart tag 2",
            #     "value": "Heart",
            #     "key": "360438485fd34264921ca19bd43b0b71"
            # }
        """

        return VolumeTag(
            meta=take_with_default(meta, self.meta),
            value=take_with_default(value, self.value),
            key=take_with_default(key, self.key),
            sly_id=take_with_default(sly_id, self.sly_id),
            labeler_login=take_with_default(labeler_login, self.labeler_login),
            updated_at=take_with_default(updated_at, self.updated_at),
            created_at=take_with_default(created_at, self.created_at),
        )



import supervisely as sly
import supervisely.io.env as sly_env
from supervisely._utils import is_production

import traceback
from rich.console import Console


def set_output_directory_run(task_id: int, team_id: int, dst_dir: str) -> bool:
    """
    Note: arguments task_id and team_id were kept for backward compatibility
    """

    console = Console()

    if is_production():

        api = sly.Api.from_env()
        task_id = sly_env.task_id()

        if api.task.get_info_by_id(task_id) is None:
            console.print(
                f"\nError: Task with ID={task_id} is either not exist or not found in your account\n",
                style="bold red",
            )
            return False

        team_id = api.task.get_info_by_id(task_id)["teamId"]

        if api.team.get_info_by_id(team_id) is None:
            console.print(
                f"\nError: Team with ID={team_id} is either not exist or not found in your account\n",
                style="bold red",
            )
            return False

    try:
        sly.output.set_directory(dst_dir)
        console.print("\nSetting task output directory succeed\n", style="bold green")
        return True

    except:
        console.print("\nSetting task output directory failed\n", style="bold red")
        traceback.print_exc()
        return False



from __future__ import annotations

from typing import List
from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


"""

l = sly.app.widgets.Text(text="left part", status="success")
items = [
    sly.app.widgets.Select.Item(label="CPU", value="cpu"),
    sly.app.widgets.Select.Item(label="GPU 0", value="cuda:0"),
    sly.app.widgets.Select.Item(value="option3"),
]
r = sly.app.widgets.Select(items=items, filterable=True, placeholder="select me")

# sidebar = sly.app.widgets.Sidebar(left_pane=l, right_pane=item)

g1_items = [
    sly.app.widgets.Menu.Item(title="m1", content=r),
    sly.app.widgets.Menu.Item(title="m2", content=l),
]
g2_items = [
    sly.app.widgets.Menu.Item(title="m3"),
    sly.app.widgets.Menu.Item(title="m4"),
]
g1 = sly.app.widgets.Menu.Group("g1", g1_items)
g2 = sly.app.widgets.Menu.Group("g2", g2_items)
menu = sly.app.widgets.Menu(groups=[g1, g2])
# menu = sly.app.widgets.Menu(items=g1_items)

"""


class Menu(Widget):
    class Item:
        def __init__(
            self,
            title: str,
            # heading: str = None,
            content: Widget = None,
            index: str = None,
            icon: str = None,
        ) -> Menu.Item:
            self.title = title
            # self.heading = title if heading is None else heading
            self.index = index
            self.content = content
            if index is None:
                self.index = str(self.title)
            self.icon = icon

        def to_json(self):
            return {
                "title": self.title,
                # "heading": self.heading,
                "index": self.index,
                "icon": self.icon,
            }

    class Group:
        def __init__(self, title: str, items: List[Menu.Item] = None) -> Menu.Group:
            self.title = title
            self.items = items

        def to_json(self):
            return {
                "title": self.title,
                "items": [item.to_json() for item in self.items],
            }

    def __init__(
        self,
        items: List[Menu.Item] = None,
        groups: List[Menu.Group] = None,
        index: str = None,
        width_percent: int = 25,
        widget_id: str = None,
    ):
        if items is None and groups is None:
            raise ValueError("One of the arguments has to be defined: items or groups")

        if items is not None and groups is not None:
            raise ValueError(
                "Only one of the arguments has to be defined: items or groups"
            )

        self._items = items
        if self._items is None:
            self._items = []
        self._groups = groups
        if self._groups is None:
            self._groups = []
        self._width_percent = width_percent
        self._options = {"sidebarWidth": self._width_percent}
        self._index = index
        if self._index is None:
            self._index = self._get_default_index()
        else:
            self._validate_index()

        StateJson()["app_body_padding"] = "0px"
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        res = {"items": [], "groups": []}
        if self._items is not None:
            res["items"] = [item.to_json() for item in self._items]
        if self._groups is not None:
            res["groups"] = [g.to_json() for g in self._groups]
        return res

    def _validate_index(self):
        for item in self._items:
            if self._index == item.index:
                return

        for group in self._groups:
            for item in group.items:
                if self._index == item.index:
                    return

        raise ValueError(
            f"Index {self._index} not found, there is no item in menu with such index"
        )

    def _get_default_index(self):
        if len(self._items) > 0:
            return self._items[0].index
        if len(self._groups) > 0:
            return self._groups[0].items[0].index

    def get_json_state(self):
        return {"menuIndex": self._index}
